@@ -15,7 +15,7 @@
 
 // You should have received a copy of the GNU General Public License along
 // with this library; see the file COPYING.  If not, write to the Free
-// Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+// Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 // USA.
 
 // As a special exception, you may use this file as part of a free software
@@ -432,8 +432,11 @@ namespace __gnu_cxx
 		static __gthread_once_t __once = __GTHREAD_ONCE_INIT;
 		__gthread_once(&__once, _S_initialize);
 	      }
-	    else
-	      _S_get_pool()._M_initialize_once(); 
+
+	    // Double check initialization. May be necessary on some
+	    // systems for proper construction when not compiling with
+	    // thread flags.
+	    _S_get_pool()._M_initialize_once(); 
 	    __init = true;
 	  }
       }
@@ -524,8 +527,11 @@ namespace __gnu_cxx
 		static __gthread_once_t __once = __GTHREAD_ONCE_INIT;
 		__gthread_once(&__once, _S_initialize);
 	      }
-	    else
-	      _S_get_pool()._M_initialize_once(); 
+
+	    // Double check initialization. May be necessary on some
+	    // systems for proper construction when not compiling with
+	    // thread flags.
+	    _S_get_pool()._M_initialize_once(); 
 	    __init = true;
 	  }
       }
@@ -620,15 +626,12 @@ namespace __gnu_cxx
 	  typedef __mt_alloc<_Tp1, pol_type> other;
 	};
 
-      __mt_alloc() throw() 
-      { __policy_type::_S_get_pool(); }
+      __mt_alloc() throw() { }
 
-      __mt_alloc(const __mt_alloc&) throw() 
-      { __policy_type::_S_get_pool(); }
+      __mt_alloc(const __mt_alloc&) throw() { }
 
       template<typename _Tp1, typename _Poolp1>
-        __mt_alloc(const __mt_alloc<_Tp1, _Poolp1>& obj) throw()  
-        { __policy_type::_S_get_pool(); }
+        __mt_alloc(const __mt_alloc<_Tp1, _Poolp1>&) throw() { }
 
       ~__mt_alloc() throw() { }
 
