@@ -1,6 +1,6 @@
 // String based streams -*- C++ -*-
 
-// Copyright (C) 1997, 1998, 1999, 2002, 2003, 2004
+// Copyright (C) 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006
 // Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
@@ -16,7 +16,7 @@
 
 // You should have received a copy of the GNU General Public License along
 // with this library; see the file COPYING.  If not, write to the Free
-// Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+// Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 // USA.
 
 // As a special exception, you may use this file as part of a free software
@@ -127,16 +127,18 @@ namespace std
       __string_type
       str() const
       {
+	__string_type __ret;
 	if (this->pptr())
 	  {
 	    // The current egptr() may not be the actual string end.
 	    if (this->pptr() > this->egptr())
-	      return __string_type(this->pbase(), this->pptr());
+	      __ret = __string_type(this->pbase(), this->pptr());
 	    else
- 	      return __string_type(this->pbase(), this->egptr());
+ 	      __ret = __string_type(this->pbase(), this->egptr());
 	  }
 	else
-	  return _M_string;
+	  __ret = _M_string;
+	return __ret;
       }
 
       /**
@@ -151,7 +153,7 @@ namespace std
       {
 	// Cannot use _M_string = __s, since v3 strings are COW.
 	_M_string.assign(__s.data(), __s.size());
-	_M_stringbuf_init(this->_M_mode);
+	_M_stringbuf_init(_M_mode);
       }
 
     protected:
@@ -159,14 +161,25 @@ namespace std
       void
       _M_stringbuf_init(ios_base::openmode __mode)
       {
-	this->_M_mode = __mode;
-
+	_M_mode = __mode;
 	__size_type __len = 0;
-	if (this->_M_mode & (ios_base::ate | ios_base::app))
+	if (_M_mode & (ios_base::ate | ios_base::app))
 	  __len = _M_string.size();
 	_M_sync(const_cast<char_type*>(_M_string.data()), 0, __len);
       }
 
+      virtual streamsize
+      showmanyc()
+      { 
+	streamsize __ret = -1;
+	if (_M_mode & ios_base::in)
+	  {
+	    _M_update_egptr();
+	    __ret = this->egptr() - this->gptr();
+	  }
+	return __ret;
+      }
+
       virtual int_type
       underflow();
 
@@ -198,10 +211,10 @@ namespace std
 	    // things will quickly blow up.
 	    
 	    // Step 1: Destroy the current internal array.
-	    _M_string.assign(__s, __n);
+	    _M_string.clear();
 	    
 	    // Step 2: Use the external array.
-	    _M_sync(__s, 0, 0);
+	    _M_sync(__s, __n, 0);
 	  }
 	return this;
       }
@@ -217,43 +230,15 @@ namespace std
       // Internal function for correctly updating the internal buffer
       // for a particular _M_string, due to initialization or
       // re-sizing of an existing _M_string.
-      // Assumes: contents of _M_string and internal buffer match exactly.
-      // __i == _M_in_cur - _M_in_beg
-      // __o == _M_out_cur - _M_out_beg
       void
-      _M_sync(char_type* __base, __size_type __i, __size_type __o)
-      {
-	const bool __testin = this->_M_mode & ios_base::in;
-	const bool __testout = this->_M_mode & ios_base::out;
-	char_type* __end = __base + _M_string.size();
-
-	if (__testin)
-	  this->setg(__base, __base + __i, __end);
-	if (__testout)
-	  {
-	    // If __base comes from setbuf we cannot trust capacity()
-	    // to match the size of the buffer area: in general, after
-	    // Step 1 above, _M_string.capacity() >= __n.
-	    if (__base == _M_string.data())
-	      this->setp(__base, __base + _M_string.capacity());
-	    else
-	      this->setp(__base, __end);
-	    this->pbump(__o);
-	    // egptr() always tracks the string end. When !__testin,
-	    // for the correct functioning of the streambuf inlines
-	    // the other get area pointers are identical.
-	    if (!__testin)
-	      this->setg(__end, __end, __end);
-	  }
-      }
+      _M_sync(char_type* __base, __size_type __i, __size_type __o);
 
       // Internal function for correctly updating egptr() to the actual
       // string end.
       void
       _M_update_egptr()
       {
-	const bool __testin = this->_M_mode & ios_base::in;
-
+	const bool __testin = _M_mode & ios_base::in;
 	if (this->pptr() && this->pptr() > this->egptr())
 	  if (__testin)
 	    this->setg(this->eback(), this->gptr(), this->pptr());
